------------------------------------------------------------------------
-- Minimum implemented using insertion sort
------------------------------------------------------------------------

open import Data.Nat
open import Data.Bool hiding (if_then_else_)
open import Thunk

-- The module is parameterised on a comparison function. For
-- simplicity it is assumed that comparisons take constant time.

module Minimum
  {a : Set}
  {c≤ : ℕ}
  (_≤_ : a -> a -> Thunk c≤ Bool)
  where

open import Data.Function
open import Relation.Binary.PropositionalEquality
open import Data.Nat.Properties
open ℕ-semiringSolver
open import Algebra
import Algebra.Operations as Ops
open Ops (CommutativeSemiring.semiring ℕ-commutativeSemiring)
open import Data.Fin
open import Data.Vec hiding (head; last; _>>=_)

open import LazyVec.Simple

------------------------------------------------------------------------
-- Boring lemmas used below

abstract

  lem₀ : c≤ + 2 ≡ 2 + c≤
  lem₀ = prove (c≤ ∷ []) (X :+ con 2) (con 2 :+ X) ≡-refl
    where X = var zero

  lem₁ : forall c -> 2 + c≤ + c + 1 ≡ 3 + c≤ + c
  lem₁ c = prove (2 + c≤ + c ∷ []) (X :+ con 1) (con 1 :+ X) ≡-refl
    where X = var zero

  lem₂ : forall c -> c + (3 + c≤) ≡ 3 + c≤ + c
  lem₂ c = prove (c ∷ 3 + c≤ ∷ []) (C :+ X) (X :+ C) ≡-refl
    where C = var zero; X = var (suc zero)

  lem₃ : forall n -> (1 + n * (4 + c≤)) + (3 + c≤) ≡ (1 + n) * (4 + c≤)
  lem₃ n = prove (n ∷ c≤ ∷ [])
                 ((con 1 :+ N :* (con 4 :+ C)) :+ (con 3 :+ C))
                 ((con 1 :+ N) :* (con 4 :+ C))
                 ≡-refl
    where N = var zero; C = var (suc zero)

  lem₄ : forall n -> (3 + c≤) + n * (3 + c≤) ≡ (1 + n) * (3 + c≤)
  lem₄ n = prove (n ∷ 3 + c≤ ∷ [])
                 (C :+ N :* C)
                 ((con 1 :+ N) :* C)
                 ≡-refl
    where N = var zero; C = var (suc zero)

  lem₅ : forall n -> 1 + (1 + n) * (4 + c≤) + 1 ≡ 2 + (4 + c≤) * (1 + n)
  lem₅ n = prove (n ∷ 4 + c≤ ∷ [])
                 (con 1 :+ (con 1 :+ N) :* C :+ con 1)
                 (con 2 :+ C :* (con 1 :+ N))
                 ≡-refl
    where N = var zero; C = var (suc zero)

  lem₆ : forall n
       -> (1 + (1 + n) * (4 + c≤)) +
          (1 + (1 + n) * (1 + (1 + n) * (3 + c≤))) ≡
          (10 + 2 * c≤) + (11 + 3 * c≤) * n + (3 + c≤) * n ^ 2
  lem₆ n = prove (n ∷ c≤ ∷ [])
                 (con 1 :+ (con 1 :+ N) :* (con 4 :+ C) :+
                  (con 1 :+
                   (con 1 :+ N) :*
                     (con 1 :+ (con 1 :+ N) :* (con 3 :+ C))))
                 ((con 10 :+ con 2 :* C) :+
                  (con 11 :+ con 3 :* C) :* N :+ (con 3 :+ C) :* N :^ 2)
                 ≡-refl
    where N = var zero; C = var (suc zero)

------------------------------------------------------------------------
-- A thunkified version of if_then_else_

infix 0 if_then_else_

if_then_else_ : forall {a} -> Bool -> a -> a -> Thunk 1 a
if true  then x else y = √ return x
if false then x else y = √ return y

------------------------------------------------------------------------
-- Insertion sort and minimum

-- Inserts element into list. If the input is ordered, then the output
-- is also ordered.

insert
  :  forall {c n}
  -> a -> VecL a c n -> Thunk (3 + c≤) (VecL a (3 + c≤ + c) (1 + n))
insert {c}         x []L       = √
  returnW (1 + c≤) (x ∷L returnW (2 + (c≤ + c)) []L)
insert {c} {suc _} x (y ∷L ys) = √ cast lem₀
  (x ≤ y >>= \b -> √
   if b then
     x ∷L cast (lem₁ c) (wait (1 + c≤ + c) (waitL (1 + c≤) (y ∷L ys)))
    else
     y ∷L cast (lem₂ c) (insert x =<< ys))

-- Sorts list (with respect to _≤_).

sort :  forall {n}
     -> Vec a n -> Thunk (1 + n * (4 + c≤)) (VecL a (n * (3 + c≤)) n)
sort         []       = √ return []L
sort {suc n} (x ∷ xs) = √
  ≡-subst (\x -> Thunk ((1 + n) * (4 + c≤)) (VecL a x (suc n))) (lem₄ n)
  (cast (lem₃ n)
     (insert x =<< sort xs))

-- Finds the minimum element in the list (if _≤_ represents a total
-- ordering).

minimum
  :  forall {n}
  -> Vec a (1 + n) -> Thunk (3 + (4 + c≤) * (1 + n)) a
minimum {n} xs = √ cast (lem₅ n) (head =<< sort xs)

-- Finds the maximum element in the list (if _≤_ represents a total
-- ordering). There are of course better ways to implement maximum...

maximum
  :  forall {n}
  -> Vec a (1 + n)
  -> Thunk ((11 + 2 * c≤) + (11 + 3 * c≤) * n + (3 + c≤) * n ^ 2) a
maximum {n} xs = √ cast (lem₆ n) (last =<< sort xs)
