-- The implementation used here is superseded by the new library. The
-- type signatures of the core functions are the same, though. Please
-- ignore this file.

-- # import nat
-- # import eq
-- # import order
-- # import monad

-- It takes at most n evaluation steps to evaluate the thunk and get
-- the a.

mutual

  Thunk : ℕ -> Set -> Set
  Thunk n a = Thunk' a n

  data Thunk' (a : Set) : ℕ -> Set where
    thunk (n : ℕ) : a -> Thunk n a

-- Note that this function must not be used in thunkified code. Its
-- purpose is to allow thunkified functions to be run in nonthunkified
-- code.

force (n : ℕ) (a : Set) : Thunk n a -> a
force (thunk x) = x

congThunk (m, n : ℕ) (a, b : Set)
  : m ≡ n -> a ≡′ b -> Thunk m a ≡′ Thunk n b
congThunk refl refl' = refl'

substThunk (m, n : ℕ) (a, b : Set)
  : m ≡ n -> a ≡′ b -> Thunk m a -> Thunk n b
substThunk refl refl' x = x

-- The implementation of (>>=) assumes that Thunk is a newtype.

infixl 1 >>=

(>>=) (a, b : Set) (m, n : ℕ)
  : Thunk m a -> (a -> Thunk n b) -> Thunk (m + n) b
c >>= f = thunk (force (f (force c)))

-- Note that we may not seq something without first paying for it. In
-- other words, we should only seq things whose type does not start
-- with Thunk. (In a language with a Haskell-style seq.)

seq (a, b : Set) : a -> b -> b
seq = const id

-- seqBind is a helper function which makes seqing more convenient.
-- Note that it is only safe to use if the type a is not a Thunk.

-- As observed by Ross, this construct can be abused:
-- .
--   pay 0 t `seqBind` \t' -> ...
-- .
-- t is seqed, but we don't pay for the work.

seqBind (a, b : Set) (m, n : ℕ)
  : Thunk m a -> (a -> Thunk n b) -> Thunk (m + n) b
m `seqBind` f = m `seq` (m >>= f)

-- We can use the following method to simulate a strict language:
-- .
-- 1. Make Thunk a lifted data type.
-- 2. Make bind strict (use strictBind, see below).
-- .
-- Why do we need to lift Thunk? Because we don't want the result of
-- pay to be forced immediately.
-- .
-- The reasoning above is clearly lacking in detail, and should be
-- formalised.

-- Assumes lifted Thunks.

strictBind (a, b : Set) (m, n : ℕ)
  : Thunk m a -> (a -> Thunk n b) -> Thunk (m + n) b
thunk c `strictBind` f = c `seq` (thunk c >>= f)

dependentBind (a, b : Set) (m : ℕ)
  :  (f : a -> ℕ)
  -> (c : Thunk m a)
  -> ((x : a) -> Thunk (f x) b)
  -> Thunk (m + f (force c)) b
dependentBind _ c g = thunk (force (g (force c)))

-- Note that dependentBind₂ can be implemented in terms of
-- dependentBind.

dependentBind₂ (a₁, a₂, b : Set) (m₁, m₂ : ℕ)
  :  (f : a₁ -> a₂ -> ℕ)
  -> (c₁ : Thunk m₁ a₁)
  -> (c₂ : Thunk m₂ a₂)
  -> ((x₁ : a₁) -> (x₂ : a₂) -> Thunk (f x₁ x₂) b)
  -> Thunk (m₁ + m₂ + f (force c₁) (force c₂)) b
dependentBind₂ {a₁} {a₂} {b} {m₁} {m₂} f c₁ c₂ g =
  substThunk addAssociative refl' $
    dependentBind f₁ c₁ (\x₁ -> dependentBind (f x₁) c₂ (g x₁))
  where
  f₁ : a₁ -> ℕ
  f₁ x₁ = m₂ + f x₁ (force c₂)

return (a : Set) : a -> Thunk 0 a
return = thunk

thunkM : AnnotatedMonad natPlus Thunk
thunkM = struct
  bind   = (>>=)
  return = return

  -- Proofs of the monad laws for this annotated monad:

  leftUnit x f = lU
    where
    lU : subst (\n -> Thunk n _) natPlus.leftUnit (return _ x >>= f) ≡ f x
    lU with f x => _ (thunk _) = refl

  rightUnit {n} {a} = rU
    where
    rU :  (c : Thunk n a)
       -> subst (\n' -> Thunk n' a) natPlus.rightUnit (c >>= return a) ≡ c
    rU (thunk _) = refl

  assoc {n₁} {n₂} {n₃} {a₁} {a₂} {a₃} c f g = ass c
    where
    stripLemma (n₁, n₂ : ℕ) (a : Set) (x, y : a)
      : thunk {n₁} x ≡ thunk {n₁} y -> thunk {n₂} x ≡ thunk {n₂} y
    stripLemma refl = refl

    lemma (n₁, n₂ : ℕ) (a₁, a₂ : Set)
      :  (h : a₁ -> Thunk n₂ a₂) -> (x : a₁) -> (y : a₂)
      -> h x ≡ thunk y
      -> (thunk {n₁} x >>= h) ≡ thunk {n₁ + n₂} y
    lemma h x y eq with inspect (h x) =>
      _ h x y eq (thunk y' :≡ eq') =
        rewrite-plugin(eq', stripLemma $ sym eq' `trans` eq)

    lemma₁
      :  (x : a₁) -> (y : a₂) -> (z : a₃)
      -> f x ≡ thunk y -> g y ≡ thunk z
      -> ((thunk x >>= f) >>= g) ≡ thunk z
    lemma₁ x y z eq₁ eq₂ =
      -- ((thunk x >>= f) >>= g) ≡ (thunk y >>= g)
      cong (\i -> i >>= g)
           -- (thunk x >>= f) ≡ thunk y
           (lemma {n₁} {n₂} f x y eq₁)
        `trans`
      -- (thunk y >>= g) ≡ thunk z
      lemma g y z eq₂

    lemma₂
      :  (x : a₁) -> (y : a₂) -> (z : a₃)
      -> f x ≡ thunk y -> g y ≡ thunk z
      -> (thunk {n₁} x >>= (\i -> f i >>= g)) ≡ thunk z
    lemma₂ x y z eq₁ eq₂ =
      -- (thunk x >>= (\i -> f i >>= g)) ≡ thunk z
      lemma (\i -> f i >>= g) x z
            -- (f x >>= g) ≡ thunk z
            ( -- (f x >>= g) ≡ (thunk y >>= g)
              cong (\i -> i >>= g) eq₁
                `trans`
              -- (thunk y >>= g) ≡ thunk z
              lemma g y z eq₂
            )

    substLemma (a : Set) (n, n' : ℕ)
      :  (x : Thunk n a) -> (y : a) -> (eq : n ≡ n')
      -> x ≡ thunk y
      -> subst (\n' -> Thunk n' a) eq x ≡ thunk y
    substLemma (thunk x') x' refl refl = refl

    lemma₂'
      :  (x : a₁) -> (y : a₂) -> (z : a₃)
      -> f x ≡ thunk y -> g y ≡ thunk z
      -> subst (\n' -> Thunk n' a₃) (natPlus.assoc {n₁} {n₂} {n₃})
           (thunk x >>= (\i -> f i >>= g)) ≡
         thunk z
    lemma₂' x y z eq₁ eq₂ =
      substLemma {_} {n₁ + (n₂ + n₃)} {(n₁ + n₂) + n₃}
        (thunk x >>= (\i -> f i >>= g))
        z
        (natPlus.assoc {n₁} {n₂} {n₃})
        (lemma₂ x y z eq₁ eq₂)

    ass :  (c : Thunk n₁ a₁)
        -> ((c >>= f) >>= g) ≡
           subst (\n' -> Thunk n' a₃) natPlus.assoc
                 (c >>= \x -> f x >>= g)
    ass (thunk x) with inspect (f x) =>
      _ _ (thunk y :≡ _) with inspect (g y) =>
        _ (thunk x) (thunk y :≡ eq₁) (thunk z :≡ eq₂) =
          lemma₁ x y z eq₁ eq₂
            `trans`
          sym (lemma₂' x y z eq₁ eq₂)

infixr 1 =<<

(=<<) (a, b : Set) (m, n : ℕ)
  : (a -> Thunk m b) -> Thunk n a -> Thunk (n + m) b
(=<<) f c = c >>= f

ap (a, b : Set) (m, n : ℕ)
  : Thunk m (a -> b) -> Thunk n a -> Thunk (m + n) b
ap cf c = thunkM.do
  f <- cf
  x <- c
  return (f x)

join (a : Set) (m, n : ℕ)
  : Thunk m (Thunk n a) -> Thunk (m + n) a
join mm = mm >>= id

(∙) (a, b, c : Set) (m, n : ℕ)
  : (b -> Thunk m c) -> (a -> Thunk n b) -> (a -> Thunk (n + m) c)
f ∙ g = \x -> g x >>= f

liftM (a, b : Set) (n : ℕ)
  : (a -> b) -> Thunk n a -> Thunk n b
liftM f t = thunkM.do
  x <- t
  return (f x)

liftM2 (a, b, c : Set) (n₁, n₂ : ℕ)
  : (a -> b -> c) -> Thunk n₁ a -> Thunk n₂ b -> Thunk (n₁ + n₂) c
liftM2 f t₁ t₂ = thunkM.do
  x₁ <- t₁
  x₂ <- t₂
  return (f x₁ x₂)

slower (a : Set) (m, n : ℕ)
  : m ≤ n -> Thunk m a -> Thunk n a
slower _ x = thunk (force x)

slower' (n, m : ℕ) (a : Set)
  : Thunk m a -> Thunk (m + n) a
slower' = slower plusIncreasingRight

return' (n : ℕ) (a : Set)
  : a -> Thunk n a
return' = slower' ∘ return

pay (a : Set) (n : ℕ)
  : (m : ℕ) -> Thunk n a -> Thunk m (Thunk (n ∸ m) a)
pay m x = thunk (thunk (force x))

-- Every recursive, thunked equation should have the form
--   f ... = step $ ...
-- or
--   f ... = step' $ ...

-- The extra natural number m in the type signature of step' (see
-- below) is included to avoid having to write
--   step $ slower' ... $ ...
-- too often.

step' (m, n : ℕ) (a : Set)
  : Thunk n a -> Thunk (n + m + 1) a
step' {m} = slower' {m + 1}

step (n : ℕ) (a : Set)
  : Thunk n a -> Thunk (n + 1) a
step = step' {0}
