------------------------------------------------------------------------
-- Calculating the result of generalised pay
------------------------------------------------------------------------

-- When gpay is executed in the thunked semantics the resulting value
-- has to be transformed to be of the right type, and the heap has to
-- be updated to ensure that the correct credit is recorded. In
-- nontrivial cases this is done by reducing the tick count of a
-- variable in the value (by modifying the context and heap; this
-- yields credit in the heap), and restoring the tick count of all
-- other uses of this variable by inserting gwait∋. This module
-- defines this operation.

module OperationalSemantics.GPay where

open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import Value
open import TypeTriple
open import OperationalSemantics.Binding
open Heap↦
open import PayCtxt hiding (_∘_)

open import Data.Function using (id; _∘_)
open import Data.Nat using (ℕ; zero; _+_; _∸_; _⊓_)
open import Data.List using (_∷_; [])
open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-sym; ≡-trans; ≡-subst; ≡-cong)
open import Data.Nat.Properties using (m⊓n+n∸m≡n; i∸k∸j+j∸k≡i+j∸k)

------------------------------------------------------------------------
-- Lemmas used below

abstract

  ⁻¹_[]-lemma :  forall C₁ C₂ {σ}
              -> C₁ ⁻¹ [ C₂ [ σ ] ] ≡ C₁ ⁻¹∘ C₂ [ σ ]
  ⁻¹_[]-lemma C₁ C₂ = ≡-trans (≡-sym (⁻¹-lemma C₁)) (⁻¹[]-lemma C₁ C₂)

  ⁻¹-const●-lemma
    :  forall C C₂ {n₁ n₂ σ₁ σ₂}
    -> C ⁻¹∘ const● (C₂ [ Thunk n₂ σ₂ ]) [ Thunk n₁ σ₁ ] ≡
       C ⁻¹∘ C₂ [ Thunk n₂ σ₂ ]
  ⁻¹-const●-lemma C C₂ =
    ≡-trans (≡-sym (⁻¹[]-lemma C _)) (⁻¹[]-lemma C _)

  ×-lemma
    :  forall C C₁ C₂ {n₁ n₂ σ₁ σ₂}
    -> C ⁻¹∘ (C₁ [ Thunk n₂ σ₁ ] ×● C₂) [ Thunk n₁ σ₂ ] ≡
       C ⁻¹∘ (C₁ ●× C₂ [ Thunk n₁ σ₂ ]) [ Thunk n₂ σ₁ ]
  ×-lemma C C₁ C₂ =
    ≡-trans (≡-sym (⁻¹[]-lemma C _)) (⁻¹[]-lemma C _)

------------------------------------------------------------------------
-- Pay results

-- This type represents the result of evaluating an application of
-- gpay. It includes a new context, heap and value (or variable), plus
-- an accompanying substitution which takes other things to the new
-- context.

record GPay (_•_ : Ctxt thunked -> Ty thunked -> Set)
            (Γ : Ctxt thunked) (C : PayCtxt)
            (n m : ℕ) (σ : Ty thunked)
            : Set where
  field
    Δ  : Ctxt thunked
    ρ  : Γ ⇒ Δ
    Σ₂ : Heap Δ
    v₂ : Δ • C [ Thunk (n ∸ m) σ ]

open GPay

-- Updates the "value" field.

update-v₂
  :  forall {Γ} {_•¹_ C¹ n¹ m¹ σ¹}
                {_•²_ C² n² m² σ²}
  -> (r : GPay _•¹_ Γ C¹ n¹ m¹ σ¹)
  -> (   Δ r •¹ C¹ [ Thunk (n¹ ∸ m¹) σ¹ ]
      -> Δ r •² C² [ Thunk (n² ∸ m²) σ² ])
  -> GPay _•²_ Γ C² n² m² σ²
update-v₂ r f = record
  { Δ  = Δ r
  ; ρ  = ρ r
  ; Σ₂ = Σ₂ r
  ; v₂ = f (v₂ r)
  }

------------------------------------------------------------------------
-- Calculating the result of gpay: variable cases

mutual

  -- The vz case of gpayResult∋ is where things really happen. Most
  -- other cases merely propagate this new information.

  -- See gpayResult below for an explanation of this function's
  -- arguments.

  gpayResult∋ :  forall C (m : ℕ) {Γ τ n σ}
              -> Heap Γ
              -> Γ ∋ τ -> τ ≡ C [ Thunk n σ ]
              -> GPay _∋_ Γ C n m σ
  gpayResult∋ C₂ m Σ₁ (gwait∋ C₁ n x₁) eq =
    gpayResult∋-gwait ● m n _ _ Σ₁ x₁
                      (complete ⟨ C₁ , _ , _ ⟩ ⟨ C₂ , _ , _ ⟩ eq)

  gpayResult∋ C m .{Γ = Γ ▻ C [ Thunk n σ ]} {n = n} {σ = σ}
              (Σ₁ ▷ʰ b) (vz {Γ = Γ}) ≡-refl = record
      -- Reduce the tick count of this variable.
    { Δ  = Γ ▻ C [ Thunk (n ∸ m) σ ]
      -- The substitution ρ applies gwait∋ to occurrences of this
      -- variable.
    ; ρ  = wk ▷ cast (gwait∋ {n₁ = n ∸ m} C (m ⊓ n) vz)
      -- The binding in the heap is paid off.
    ; Σ₂ = Σ₁ ▷ʰ gpay↦ C m b
    ; v₂ = vz
    }
    where
    cast = ≡-subst (_∋_ _) (≡-cong (\i -> C [ Thunk i σ ])
                                   (m⊓n+n∸m≡n m n))

  gpayResult∋ C m (Σ₁ ▷ʰ b) (vs {σ = σ} x₁) ≡-refl = record
    { Δ  = Δ r ▻ σ
    ; ρ  = ρ r ↑
      -- Note that the substitution is applied to all outer bindings.
    ; Σ₂ = Σ₂ r ▷ʰ b /↦ ρ r
    ; v₂ = vs (v₂ r)
    }
    where r = gpayResult∋ C m Σ₁ x₁ ≡-refl

  -- This function takes care of propagating the information through
  -- applications of gwait∋.

  gpayResult∋-gwait
    :  forall C {C₁ C₂} (m n n₁ n₂ : ℕ) {Γ σ₁ σ₂} (Σ₁ : Heap Γ)
    -> (x₁ : Γ ∋ C ⁻¹∘ C₁ [ Thunk n₁ σ₁ ])
    -> (eq : ⟨ C₁ , n + n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩)
    -> GPay _∋_ Γ (C ⁻¹∘ C₂) n₂ m σ₂
  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜c- {C = C₂}) =
    update-v₂ (gpayResult∋ (C ⁻¹∘ C₂) m Σ₁ x₁ (⁻¹-const●-lemma C C₂))
              (gwait∋ (C ⁻¹∘ C₂) zero)

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜-c {C = C₁} {σ₁ = σ₁}) =
    update-v₂ (gpayResult∋ (C ⁻¹∘ const● (C₁ [ Thunk n₁ σ₁ ]))
                           m {n = n₂} {σ = σ₁} Σ₁ x₁
                           (≡-sym (⁻¹-const●-lemma C C₁)))
              (cast₁ ∘ gwait∋ (C ⁻¹∘ C₁) n ∘ cast₂)
    where
    cast₁ = ≡-subst (_∋_ _) (≡-sym (⁻¹-const●-lemma C C₁))
    cast₂ = ≡-subst (_∋_ _) (⁻¹-const●-lemma C C₁)

  gpayResult∋-gwait C m n n₁ .(n + n₁) Σ₁ x₁ (≜● {σ = σ}) =
    update-v₂ r (cast ∘ gwait∋ (C ⁻¹) (n ∸ (m ∸ n₁)))
    where
    r    = gpayResult∋ (C ⁻¹) m Σ₁ x₁ ≡-refl
    cast = ≡-subst (_∋_ _) (≡-cong (\i -> C ⁻¹ [ Thunk i σ ])
                                   (i∸k∸j+j∸k≡i+j∸k n n₁ m))

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜●T {C = C'}) =
    update-v₂ (gpayResult∋ (C ⁻¹∘ Thunk● n₁ C') m Σ₁ x₁
                           (⁻¹_[]-lemma C (Thunk● n₁ C')))
              (cast₁ ∘ gwait∋ (C ⁻¹) n ∘ cast₂)
    where
    cast₁ = ≡-subst (_∋_ _) (⁻¹_[]-lemma C (Thunk● (n + n₁) C'))
    cast₂ = ≡-subst (_∋_ _) (≡-sym (⁻¹_[]-lemma C (Thunk● n₁ C')))

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜T● {C = C'}) =
    update-v₂ (gpayResult∋ (C ⁻¹) m Σ₁ x₁
                           (≡-sym (⁻¹_[]-lemma C (Thunk● n₂ C'))))
              (cast₁ ∘ gwait∋ (C ⁻¹∘ Thunk● (n₂ ∸ m) C') n ∘ cast₂)
    where
    cast₁ = ≡-subst (_∋_ _) (≡-sym (⁻¹_[]-lemma C (Thunk● (n₂ ∸ m) C')))
    cast₂ = ≡-subst (_∋_ _) (⁻¹_[]-lemma C (Thunk● (n₂ ∸ m) C'))

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜TT {n = n'} eq) =
    gpayResult∋-gwait (Thunk● n' C) m n n₁ n₂ Σ₁ x₁ eq

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜●×² {σ = σ} eq) =
    gpayResult∋-gwait (C ●× σ) m n n₁ n₂ Σ₁ x₁ eq

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁ (≜×●² {σ = σ} eq) =
    gpayResult∋-gwait (σ ×● C) m n n₁ n₂ Σ₁ x₁ eq

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁
                    (≜●××● {C₁ = C₁} {C₂ = C₂} {σ₁ = σ₁} {σ₂ = σ₂}) =
    update-v₂ (gpayResult∋ (C ⁻¹∘ (C₁ [ Thunk n₁ σ₁ ] ×● C₂)) m Σ₁ x₁
                           (≡-sym (×-lemma C C₁ C₂)))
              (cast₁ ∘ gwait∋ (C ⁻¹∘ (C₁ ●× C₂ [ Thunk (n₂ ∸ m) σ₂ ])) n
                     ∘ cast₂)
    where
    cast₁ = ≡-subst (_∋_ _) (≡-sym (×-lemma C C₁ C₂))
    cast₂ = ≡-subst (_∋_ _) (×-lemma C C₁ C₂)

  gpayResult∋-gwait C m n n₁ n₂ Σ₁ x₁
                    (≜×●●× {C₁ = C₁} {C₂ = C₂} {σ₁ = σ₁} {σ₂ = σ₂}) =
    update-v₂ (gpayResult∋ (C ⁻¹∘ (C₁ ●× C₂ [ Thunk n₁ σ₂ ])) m Σ₁ x₁
                           (×-lemma C C₁ C₂))
              (cast₁ ∘ gwait∋ (C ⁻¹∘ (C₁ [ Thunk (n₂ ∸ m) σ₁ ] ×● C₂)) n
                     ∘ cast₂)
    where
    cast₁ = ≡-subst (_∋_ _) (×-lemma C C₁ C₂)
    cast₂ = ≡-subst (_∋_ _) (≡-sym (×-lemma C C₁ C₂))

------------------------------------------------------------------------
-- Calculating the result of gpay

-- This function calculates the result of evaluating gpay C m e in the
-- heap Σ, given the heap and value which are the result of evaluating
-- e in Σ. The functions above are used when a variable inside the
-- value is paid off.

gpayResult :  forall {Γ} C m (Σ₁ : Heap Γ) {n σ}
           -> Γ ⊢⇓ C [ Thunk n σ ]
           -> GPay _⊢⇓_ Γ C n m σ
gpayResult ● m Σ₁ (return⋆⇓ v₁) = record
  { Δ  = _
  ; ρ  = idˢ
  ; Σ₂ = Σ₁
  ; v₂ = return⋆⇓ v₁
  }

gpayResult (const● σ) m Σ₁ v₁ = record
  { Δ  = _
  ; ρ  = idˢ
  ; Σ₂ = Σ₁
  ; v₂ = v₁
  }

gpayResult (C ●× σ₂) m Σ₁ (x₁ ,⇓ y₁) =
  update-v₂ r (\x₂ -> (x₂ ,⇓ (y₁ /∋ ρ r)))
  where r = gpayResult∋ C m Σ₁ x₁ ≡-refl

gpayResult (σ₁ ×● C) m Σ₁ (x₁ ,⇓ y₁) =
  update-v₂ r (\y₂ -> ((x₁ /∋ ρ r) ,⇓ y₂))
  where r = gpayResult∋ C m Σ₁ y₁ ≡-refl

gpayResult (Thunk● k C) m Σ₁ (return⋆⇓ v₁) =
  update-v₂ (gpayResult C m Σ₁ v₁) return⋆⇓
