------------------------------------------------------------------------
-- The type system of a simply typed λ-calculus with products and
-- natural numbers, and perhaps also thunks
------------------------------------------------------------------------

module TypeSystem where

open import Kind

open import Data.Nat using (ℕ; zero; suc; _+_; _∸_)

infixl 70 _·_
infixr 60 _,_
infix  50 λ_ λ²_
infixl 46 _>>=_
infix  43 √_
infix  41 _[_]
infixr 40 _×_ _●×_ _×●_
infixr 30 _→_
infixl 20 _▻_
infix  10 _∋_ _⊢_

------------------------------------------------------------------------
-- Types

-- Types, indexed on their kind (simple or thunked).

data Ty : Kind -> Set where
  Nat   : forall {k} -> Ty k
  _×_   : forall {k} -> Ty k -> Ty k -> Ty k
  _→_   : forall {k} -> Ty k -> Ty k -> Ty k
  Thunk : ℕ -> Ty thunked -> Ty thunked

-- Thunk? n τ is Thunk n τ in the thunked case, but simplifies to τ in
-- the simple case. (Note that Maybe ℕ k simplifies to ⊤ when k =
-- simple, so the superfluous argument n can be inferred automatically
-- in this case.) In general, functions with names ending in a
-- question mark "vanish" in the simple case.

Thunk? : forall {k} -> Maybe ℕ k -> Ty k -> Ty k
Thunk? {k = simple}  n τ = τ
Thunk? {k = thunked} n τ = Thunk n τ

------------------------------------------------------------------------
-- Contexts

-- Contexts used by generalised pay. (Here context is used in the
-- sense of something with holes, not in the sense of a list of
-- types.)

data PayCtxt : Set where
  ●      : PayCtxt
  const● : Ty thunked -> PayCtxt
  _●×_   : PayCtxt -> Ty thunked -> PayCtxt
  _×●_   : Ty thunked -> PayCtxt -> PayCtxt
  Thunk● : ℕ -> PayCtxt -> PayCtxt

-- Fills the hole in a PayCtxt.

_[_] : PayCtxt -> Ty thunked -> Ty thunked
●          [ τ ] = τ
const● σ   [ τ ] = σ
(C ●× σ)   [ τ ] = C [ τ ] × σ
(σ ×● C)   [ τ ] = σ × C [ τ ]
Thunk● n C [ τ ] = Thunk n (C [ τ ])

------------------------------------------------------------------------
-- Variables

-- Contexts used to keep track of variables.

data Ctxt (k : Kind) : Set where
  -- Empty context.
  ε   : Ctxt k
  -- Context extension.
  _▻_ : Ctxt k -> Ty k -> Ctxt k

-- Type signature for term-like things.

TmLike : Set1
TmLike = forall {k} -> Ctxt k -> Ty k -> Set

-- Variables (de Bruijn indices).

data _∋_ : TmLike where
  -- Variable zero.
  vz : forall {k} {Γ : Ctxt k} {σ}   -> Γ ▻ σ ∋ σ
  -- Variable successor.
  vs : forall {k} {Γ : Ctxt k} {σ τ} -> Γ ∋ τ -> Γ ▻ σ ∋ τ
  -- Pretend that a computation takes more steps than it actually
  -- does. (Note that this constructor only exists in the thunked
  -- setting.)
  gwait∋ :  forall {Γ σ n₁} C n₂
         -> Γ ∋ C [ Thunk n₁ σ ]
         -> Γ ∋ C [ Thunk (n₂ + n₁) σ ]

-- The gwait∋ constructor ("generalised waiting") is included to aid
-- the analysis of code involving generalised pay (gpay). There is
-- also a corresponding term constructor gwait. These constructors do
-- not correspond to any library primitives. In light of the
-- formalisation it would be sound to include gwait in the library,
-- though.

------------------------------------------------------------------------
-- Terms

-- The term type encodes both simple terms and thunked, run-time
-- terms, depending on the value of the Kind index. The necessary uses
-- of ticks in the run-time terms have been encoded using types. As an
-- example, consider λ_:
--
-- * In the simple case the type of λ_ simplifies to the following
--   type, equivalent to the one used in the paper:
--
--     λ_ :  forall {Γ : Ctxt simple} {n : ⊤} {σ τ}
--        -> Γ ▻ σ ⊢ τ -> Γ ⊢ σ → τ
--
--   (⊤ is the unit type, so the n could be removed. ⊤ also has η
--   equality, so the unique value of n never needs to be written
--   explicitly; it can always be inferred.)
--
-- * In the thunked case the type of λ_ instead simplifies to:
--
--     λ_ :  forall {Γ : Ctxt thunked} {n : ℕ} {σ τ}
--        -> Γ ▻ σ ⊢ Thunk n τ -> Γ ⊢ σ → Thunk (suc n) τ
--
--   The run-time form of lambda is λx. √ e. Here e has to have type
--   Thunk n τ for some n and τ; this is reflected in the type of λ_.
--   The type of the entire expression, with the extra tick, then
--   becomes σ → Thunk (suc n) τ (where σ is the type of x); this is
--   also reflected in the type of λ_.

-- Some extensions, mentioned in the paper, are included in the term
-- type. To show that partial applications are OK the constructor s
-- can be partially applied, and I have also included a two-argument
-- lambda. Furthermore general fixpoints are allowed (fix) and the
-- generalised pay (gpay) is included.

data _⊢_ : TmLike where
  -- Lambda calculus.
  var  :  forall {k} {Γ : Ctxt k} {τ}
       -> Γ ∋ τ -> Γ ⊢ τ
  λ_   :  forall {k} {Γ : Ctxt k} {n σ τ}
       -> Γ ▻ σ ⊢ Thunk? n τ -> Γ ⊢ σ → Thunk? (suc' k n) τ
  λ²_  :  forall {k} {Γ : Ctxt k} {n σ₁ σ₂ τ}
       -> Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk? n τ
       -> Γ ⊢ σ₁ → σ₂ → Thunk? (suc' k n) τ

  _·_  :  forall {k} {Γ : Ctxt k} {σ τ}
       -> Γ ⊢ σ → τ -> Γ ⊢ σ -> Γ ⊢ τ
  fix  :  forall {k} {Γ : Ctxt k} {n σ}
       -> Γ ▻ Thunk? (suc' k n) σ ⊢ Thunk? n σ
       -> Γ ⊢ Thunk? (suc' k n) σ

  -- Natural numbers.
  z      :  forall {k} {Γ : Ctxt k}
         -> Γ ⊢ Nat
  s      :  forall {k} {Γ : Ctxt k}
         -> Γ ⊢ Nat → Nat
  natrec :  forall {k} {Γ : Ctxt k} {n σ}
         -> Γ ⊢ Thunk? n σ
         -> Γ ▻ Nat ▻ Thunk? (suc' k n) σ ⊢ Thunk? n σ
         -> Γ ⊢ Nat → Thunk? (suc' k n) σ

  -- Products.
  _,_     :  forall {k} {Γ : Ctxt k} {σ τ}
          -> Γ ⊢ σ -> Γ ⊢ τ -> Γ ⊢ σ × τ
  uncurry :  forall {k} {Γ : Ctxt k} {n σ₁ σ₂ τ}
          -> Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk? n τ
          -> Γ ⊢ σ₁ × σ₂ → Thunk? (suc' k n) τ

  -- The following constructors only exist in the thunked case (by
  -- virtue of the types assigned to them). Note that force is not
  -- included, since only run-time terms are considered.

  -- Thunk library primitives.
  return :  forall {Γ σ} -> Γ ⊢ σ -> Γ ⊢ Thunk zero σ
  _>>=_  :  forall {Γ σ τ n₁ n₂}
         -> Γ ⊢ Thunk n₁ σ
         -> Γ ⊢ σ → Thunk n₂ τ
         -> Γ ⊢ Thunk (n₁ + n₂) τ
  √_     :  forall {Γ σ n}
         -> Γ ⊢ Thunk n σ -> Γ ⊢ Thunk (suc n) σ
  pay    :  forall {Γ σ n₁} n₂
         -> Γ ⊢ Thunk n₁ σ
         -> Γ ⊢ Thunk n₂ (Thunk (n₁ ∸ n₂) σ)
  gpay   :  forall {Γ σ n₁} C n₂
         -> Γ ⊢ C [ Thunk n₁ σ ]
         -> Γ ⊢ Thunk n₂ (C [ Thunk (n₁ ∸ n₂) σ ])
  gwait  :  forall {Γ σ n₁} C n₂
         -> Γ ⊢ C [ Thunk n₁ σ ]
         -> Γ ⊢ C [ Thunk (n₂ + n₁) σ ]

-- The term type includes both a simple pay and a generalised one
-- (gpay). The function pay is a special case of gpay, so it could be
-- omitted, but this arrangement makes it relatively easy to see (some
-- of) the additional complexity incurred by the addition of gpay.
