------------------------------------------------------------------------
-- Values
------------------------------------------------------------------------

module Value where

open import Kind
open import TypeSystem
open import Erasure
open import Substitution
open VarSubst

open import Data.Nat using (zero; suc; _+_)

infixr 60 _,⇓_
infix  50 λ⇓_ λ²⇓_
infixl 48 _/⊢⇓_
infix  10 _⊢⇓_

-- Values. A value is the result of evaluating a term in a "full" heap
-- (where all variables are bound to something) to WHNF. Whenever
-- possible these constructors take variables as arguments, to ensure
-- full sharing if they are evaluated further.

-- Just as before simple values and thunked, run-time values share the
-- same definition.

data _⊢⇓_ : TmLike where
  -- Lambda.
  λ⇓_  :  forall {k} {Γ : Ctxt k} {σ τ n}
       -> Γ ▻ σ ⊢ Thunk? n τ
       -> Γ ⊢⇓ σ → Thunk? (suc' k n) τ
  λ²⇓_ :  forall {k} {Γ : Ctxt k} {σ₁ σ₂ τ n}
       -> Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk? n τ
       -> Γ ⊢⇓ σ₁ → σ₂ → Thunk? (suc' k n) τ

  -- Natural numbers. An unapplied s (the successor constructor) is
  -- represented by s⇓, while s₁⇓ is used when s has been applied to
  -- something.
  z⇓      :  forall {k} {Γ : Ctxt k} -> Γ ⊢⇓ Nat
  s⇓      :  forall {k} {Γ : Ctxt k} -> Γ ⊢⇓ Nat → Nat
  s₁⇓     :  forall {k} {Γ : Ctxt k} -> Γ ∋ Nat -> Γ ⊢⇓ Nat
  natrec⇓ :  forall {k} {Γ : Ctxt k} {σ n}
          -> Γ ∋ Thunk? n σ
          -> Γ ▻ Nat ▻ Thunk? (suc' k n) σ ⊢ Thunk? n σ
          -> Γ ⊢⇓ Nat → Thunk? (suc' k n) σ

  -- Products.
  _,⇓_     :  forall {k} {Γ : Ctxt k} {σ τ}
           -> Γ ∋ σ -> Γ ∋ τ -> Γ ⊢⇓ σ × τ
  uncurry⇓ :  forall {k} {Γ : Ctxt k} {σ₁ σ₂ τ n}
           -> Γ ▻ σ₁ ▻ σ₂ ⊢ Thunk? n τ
           -> Γ ⊢⇓ σ₁ × σ₂ → Thunk? (suc' k n) τ

  -- Thunks.
  return⋆⇓ : forall {Γ σ n} -> Γ ⊢⇓ σ -> Γ ⊢⇓ Thunk n σ

-- gwait for values. This definition takes advantage of the fact that
-- variables can be annotated using gwait∋.

gwait⇓ :  forall {Γ σ n₁} C n₂
       -> Γ ⊢⇓ C [ Thunk n₁ σ ]
       -> Γ ⊢⇓ C [ Thunk (n₂ + n₁) σ ]
gwait⇓ ●            m (return⋆⇓ v) = return⋆⇓ v
gwait⇓ (const● σ)   m v            = v
gwait⇓ (C ●× σ)     m (x ,⇓ y)     = (gwait∋ C m x ,⇓ y)
gwait⇓ (σ ×● C)     m (x ,⇓ y)     = (x ,⇓ gwait∋ C m y)
gwait⇓ (Thunk● n C) m (return⋆⇓ v) = return⋆⇓ (gwait⇓ C m v)

-- return⋆ for terms.

return⋆ : forall {Γ τ n} -> Γ ⊢ τ -> Γ ⊢ Thunk n τ
return⋆ {n = zero}  e = return e
return⋆ {n = suc n} e = √ return⋆ e

-- Converts a value to the corresponding term.

valToTm : forall {k} {Γ : Ctxt k} {τ} -> Γ ⊢⇓ τ -> Γ ⊢ τ
valToTm (λ⇓ t)          = λ t
valToTm (λ²⇓ t)         = λ² t
valToTm z⇓              = z
valToTm s⇓              = s
valToTm (s₁⇓ x)         = s · var x
valToTm (natrec⇓ x₁ t₂) = natrec (var x₁) t₂
valToTm (x₁ ,⇓ x₂)      = (var x₁ , var x₂)
valToTm (uncurry⇓ t)    = uncurry t
valToTm (return⋆⇓ v)    = return⋆ (valToTm v)

-- Variable substitutions can be applied to values.

_/⊢⇓_ : Applier _⊢⇓_ _⊢⇓_
λ⇓ t          /⊢⇓ ρ = λ⇓ (t /⊢ ρ ↑)
λ²⇓ t         /⊢⇓ ρ = λ²⇓ (t /⊢ ρ ↑ ↑)
z⇓            /⊢⇓ ρ = z⇓
s⇓            /⊢⇓ ρ = s⇓
s₁⇓ x         /⊢⇓ ρ = s₁⇓ (x /∋ ρ)
natrec⇓ x₁ t₂ /⊢⇓ ρ = natrec⇓ (x₁ /∋ ρ) (t₂ /⊢ ρ ↑ ↑)
(x₁ ,⇓ x₂)    /⊢⇓ ρ = (x₁ /∋ ρ) ,⇓ (x₂ /∋ ρ)
uncurry⇓ t    /⊢⇓ ρ = uncurry⇓ (t /⊢ ρ ↑ ↑)
return⋆⇓ v    /⊢⇓ ρ = return⋆⇓ (v /⊢⇓ ρ)

-- Erases the value.

⌜_⌝⇓ : TmEraser _⊢⇓_
⌜ λ⇓ e ⌝⇓          = λ⇓ ⌜ e ⌝
⌜ λ²⇓ e ⌝⇓         = λ²⇓ ⌜ e ⌝
⌜ z⇓ ⌝⇓            = z⇓
⌜ s⇓ ⌝⇓            = s⇓
⌜ s₁⇓ x ⌝⇓         = s₁⇓ ⌜ x ⌝∋
⌜ natrec⇓ x₁ e₂ ⌝⇓ = natrec⇓ ⌜ x₁ ⌝∋ ⌜ e₂ ⌝
⌜ x ,⇓ y ⌝⇓        = ⌜ x ⌝∋ ,⇓ ⌜ y ⌝∋
⌜ uncurry⇓ e ⌝⇓    = uncurry⇓ ⌜ e ⌝
⌜ return⋆⇓ v ⌝⇓    = ⌜ v ⌝⇓
